namespace GossamerVR.MorphFilterPlus
{    
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using UnityEngine;
    using UnityEngine.UI;
    using SimpleJSON;

    public class MorphFilterPlus : MVRScript
    {
        // strings are dangerous. It's easy to misspell a string, and then 
        // you're spending hours trying to figure out what went wrong.
        // so we put our strings all in one place, defined in advance, up here.
        // By making them fields and constants instead of hand-written strings, the compiler
        // will alert us if we misspell anything when we use them, AND we
        // only have to correct or change the string in a single place.
        ////////////////////// STRINGS ///////////////////////////////
     #region strings

        /// <summary>
        /// when we check the type string of an atom, this is what we're looking for.
        /// </summary>
        private const string k_Person = "Person";

        private const string k_morphUIPathFemale = "MorphAutoUIFemale";
        private const string k_morphUIPathMale = "MorphAutoUIMale";

        // these are the transform paths to the filter input we're going to change.
        // there are two UIs (male & female) and an alt for each, because 
        // when you first select the person, VAM renames some UI bits-- thus, the path changes.
        // when we go to find the field we need to account for both paths of both UIs,
        // because we don't know if you've already selected the person or not.
        private const string k_inputFieldPathFemale = "UI/UIPlaceHolderModel/UIModel/Canvas/Panel/Content/TabMorphsFemale/FilterInputField";
        private const string k_inputFieldPathFemaleAlt = "UI/UIPlaceHolderModel/UIModel/Canvas/Panel/Content/Female Morphs/FilterInputField";
        private const string k_inputFieldPathMale = "UI/UIPlaceHolderModel/UIModel/Canvas/Panel/Content/TabMorphsMale/FilterInputField";
        private const string k_inputFieldPathMaleAlt = "UI/UIPlaceHolderModel/UIModel/Canvas/Panel/Content/Male Morphs/FilterInputField";


        //public enum MorphBankPath
        //{
        //    FemaleChar = 0,
        //    FemaleGen,                // DO NOT USE ENUM
        //    MaleChar,                 // VAM CRASHES ON COMPILE
        //    MaleGen,                  
        //    MaleGen2
        //}

        /// <summary>
        /// the array of morph banks collectively storing all the character morphs.
        /// we store them as an array so we can foreach() them easily. Their order
        /// is important because we'll use the MorphBankPath enum as indices.
        /// </summary>
        private static readonly string[] morphBankPaths = new[]
        {
            "rescale2/geometry/MorphBanks/FemaleCharacterMorphBank(Clone)",
            "rescale2/geometry/MorphBanks/FemaleGenMorphBank(Clone)",
            "rescale2/geometry/MorphBanks/MaleCharacterMorphBank(Clone)",
            "rescale2/geometry/MorphBanks/MaleGenMorphBank(Clone)",
            "rescale2/geometry/MorphBanks/MaleGen2MorphBank(Clone)",
        };



        // while most of the time our banks and paths are iterated over,
        // we are going to want to be able to access them specifically sometimes.
        // Rather than memorizing what each int means, we use constants.
        private class BankIndex
        {
            public const int FemaleChar = 0;
            public const int FemaleGen = 1;
            public const int MaleChar = 2;
            public const int MaleGen = 3;
            public const int MaleGen2 = 4;
        }
        // this is where I would be using an enum.
        //
        //                               IF I HAD ONE


        #endregion strings
        //////////////////////////////////////////////////////////////

        /// <summary>
        /// to let people know it's working we'll turn the input fields blue.
        /// </summary>
        private static readonly Color mintyFreshColor = new Color(0.85f, 0.95f, 1f, 1f);

        /// <summary>        
        /// when true, the filter will behave as normal: applying every time
        /// the input field changes. When it's false, the filter only gets applied
        /// when the user clicks away from the field or presses enter.
        /// We show this in the plugin's UI
        /// </summary>
        private JSONStorableBool liveUpdateBoolJSON;
        
        /// <summary>
        /// A table of the people we have upgraded and their upgrades,
        /// which store references to old and new filters. 
        /// </summary>
        /// <remarks>
        /// We need this to find what person the current filter is being applied to,
        /// and so we can revert the changes when the plugin or person is removed.
        /// We also use this to make sure we don't apply an upgrade to the same person twice,
        /// and to check whether or not we should cancel an upgrade we STARTED to apply
        /// but then stopped due to removing the person or plugin. All-around important.
        /// </remarks>
        private Dictionary<Atom, UpgradeProfile> upgrades = new Dictionary<Atom, UpgradeProfile>();

        /// <summary>
        /// ongoing coroutine tasks to move through in update
        /// because I prefer having direct control instead of using StartCoroutine
        /// </summary>
        private List<IEnumerator> tasks = new List<IEnumerator>();

        private bool isInitialized = false;
        private bool isSubscribed = false;

        /// <summary>
        /// When this string is populated, it means something went wrong.
        /// it will be logged and will tell the plugin to shut down.
        /// it allows us to gracefully disable when there's an error,
        /// and not spam the user
        /// </summary>
        private string generalError = "";

        private bool hasError => this.everythingIsTerrible || !string.IsNullOrEmpty(generalError);

        /// <summary>
        /// This flag is set when we self-disable due to critical errors,        
        /// it allows us to sleep, essentially, until reset.
        /// </summary>
        private bool everythingIsTerrible = false;

        
        /// <summary>
        /// We apply filters as a coroutine when the inputfield is submitted,
        /// to allow the clear button a chance to clear it.
        /// Only used when live update is false.
        /// </summary>
        private IEnumerator submissionTask = null;

        /// <summary>
        /// all the morph uids that we will never show to the user.
        /// we populate this the very first time we upgrade a person.
        /// </summary>
        /// <remarks>
        /// this plugin works by using the 'visible' variable of morphs,
        /// which seems only used to hide morphs that should never be 
        /// visible to the user. We want to keep track of those, so we can
        /// both keep them hidden and revert properly when removed,
        /// so we do that here.
        /// </remarks>
        private HashSet<string> ignoredMorphs = null;

        /// <summary>
        /// Re-subscribes to scene events if we are re-activating post-Initialization,
        /// and also scans for any new people.
        /// </summary>
        private void OnEnable()
        {
            
            // the user might be re-enabling this after it shut down from an issue
            // so lets say there are no issues, so we can try again.
            this.everythingIsTerrible = false;
            this.generalError = "";

            // we don't want to do anything else if we're just enabling due to being
            // newly added, because VAM might have important things to do--
            // it's why MVRScript has an Init() method for initialization. 

            // plus, since this might be automatically loaded as a default session plugin,
            // I have no real way of knowing if this is getting called before or after
            // other important things are set up.

            // the solution to this is checking that we've already initialized, 
            // but that we've become unsubscribed due to OnDisable.
            // that will mean that we are re-enabling, and should re-subscribe and re-scan.
            if (this.isInitialized && !this.isSubscribed)
            {                
                this.Subscribe();
                this.ScanForPeople();
            }            
        }

        /// <summary>
        /// Unsubscribes from scene events and reverts any upgrades to people we had.
        /// called from the MonoBehaviour api, when you click the enabled checkmark in the plugin ui,
        /// or in our Sulk() method if something breaks.
        /// </summary>
        private void OnDisable()
        {            
            this.Unsubscribe();
            foreach(var pair in this.upgrades)
            {
                if (pair.Key != null) this.Log($"MorphFilterPlus reverting person '{pair.Key.name}'");
                this.RestoreMorphs(pair.Key);
                pair.Value?.Revert();
            }
            this.upgrades.Clear();
            this.tasks.Clear();            
        }

        /// <summary>
        /// Initializes by subscribing to important scene events and 
        /// scanning for people already in the current scene. 
        /// This is part of the MVRScript API and is called automatically
        /// when VAM says it's time for the plugin to initialize.
        /// </summary>
        public override void Init()
        {            
            base.Init();
            if (!this.isInitialized)
            {
                this.Subscribe();
                this.ScanForPeople();
                this.isInitialized = true;                
            }
        }

        /// <summary>
        /// Initializes the plugin's UI. Some people just use Init() but 
        /// the fact this exists makes me think it's better to do it in here.
        /// We only have one ui element right now, the live filter toggle.
        /// </summary>
        public override void InitUI()
        {
            base.InitUI();
            this.liveUpdateBoolJSON = new JSONStorableBool("Live Filtering", true);
            this.RegisterBool(this.liveUpdateBoolJSON);
            this.CreateToggle(this.liveUpdateBoolJSON);
        }

        /// <summary>
        /// Starts subscription to VAM's atom and scene events.
        /// these events will let us respond to loading a scene or atom.
        /// We call this on Init() and on OnEnable()
        /// </summary>
        private void Subscribe()
        {
            if (!this.isSubscribed)
            {
                SuperController.singleton.onAtomAddedHandlers += this.OnAtomAdded;
                SuperController.singleton.onAtomRemovedHandlers += this.OnAtomRemoved;
                SuperController.singleton.onSceneLoadedHandlers += this.OnSceneLoaded;
                SuperController.singleton.onSubSceneLoadedHandlers += this.OnSubSceneLoaded;
                this.isSubscribed = true;
            }
        }

        /// <summary>
        /// Ends subscription to VAM's atom and scene events.
        /// we unsubscribe on disable.
        /// </summary>
        private void Unsubscribe()
        {
            // since we unsub on disable, this gets called when we are quitting.
            // for all we know at that point the supercontroller is destroyed,
            // so we check first to avoid a null reference.
            // (We might be quitting but we can still be nice about it)
            if (SuperController.singleton != null)
            {                
                SuperController.singleton.onAtomAddedHandlers -= this.OnAtomAdded;
                SuperController.singleton.onAtomRemovedHandlers -= this.OnAtomRemoved;
                SuperController.singleton.onSceneLoadedHandlers -= this.OnSceneLoaded;
                SuperController.singleton.onSubSceneLoadedHandlers -= this.OnSubSceneLoaded;
                this.isSubscribed = false;
            }
            
        }


        /// <summary>
        /// Iterates the Scene's Atoms to see if there are any new people.
        /// we call this when we don't know if there are, like when
        /// becoming enabled, or when loading a new scene or subscene.        
        /// </summary>
        public void ScanForPeople()
        {
            if (!this.everythingIsTerrible)
            {
                this.Log("MorphFilterPlus.ScanForPeople()...");

                // GetSceneAtoms() is built into VAM's MVRScript API, 
                // and fetches the scene's root objects from the Supercontroller.
                // it's nice and fast, so we iterate that for people instead of 
                // using a FindObjectOfType<> or GetComponentsInChildren<> which are slower.

                foreach (var atom in this.GetSceneAtoms())
                {
                    if (atom.type == k_Person)
                    {
                        this.UpgradePerson(atom);
                    }
                }
            }
        }


        /// <summary>
        /// Begins upgrading the person's input fields as a coroutine.
        /// We only make it a coroutine so we can yield a couple frames,
        /// to ensure the person is fully prepared for filter greatness.
        /// </summary>        
        public void UpgradePerson(Atom person)
        {           
            // we only upgrade people that don't already have upgrades.
            // also we don't do it if stuff broke causing everything to be terrible.
            if (!this.upgrades.ContainsKey(person) && !this.everythingIsTerrible)
            {
                this.Log($"MorphFilterPlus upgrading a new person named '{person.name}'");

                this.upgrades[person] = new UpgradeProfile();                
                this.tasks.Add(UpgradePersonProcess(person));
            }
        }
        
        /// <summary>
        /// This is the main process that results in 
        /// converting a standard filter input to our MorphFilterPlus input.
        /// we add it to our task list via InitPerson() when we find a new one.
        /// </summary>
        private IEnumerator UpgradePersonProcess(Atom person)
        {
           
            // now we start by waiting a few frames for the character to get settled in.
            // this is ugly but it's safe, so i'ma do it.
            // I don't even know if I really need to. I'm a radical.
            // radically cautious.
            yield return null;
            yield return null;
            yield return null;


            // ANYTHING could have happened in those 3 frames. Maybe our person's destroyed now.
            // we check that they exist and that we still care about them before continuing.
            if (person != null && this.upgrades.ContainsKey(person) && !this.everythingIsTerrible)
            {
                UpgradeProfile upgrade = this.upgrades[person];
                
                // time to prepare for the magic. We fetch the person's male and female morph input fields,
                // using whatever named transform hierarchy path is accurate.
                Transform femaleInputField = person.transform.Find(k_inputFieldPathFemale);
                if (femaleInputField == null) femaleInputField = person.transform.Find(k_inputFieldPathFemaleAlt);

                Transform maleInputField = person.transform.Find(k_inputFieldPathMale);
                if (maleInputField == null) maleInputField = person.transform.Find(k_inputFieldPathMaleAlt);

                // schooled C# developers may note that i didn't use null propogation above.
                // schooled Unity developers will tell you that unity's Objects will LIE if you do. the Objects are LIARS

                
                // with InputFields in hand, we convert them to better ones.
                // our trick is duplicating the existing fields, disabling the old ones,
                // and having the new ones point events at us, so we intercept the input.
                // keeping the old ones around makes it easy to revert when 
                // either this plugin or the person is removed.
                if (femaleInputField != null && maleInputField != null)
                {                                         
                    
                    upgrade.OldFieldMale = maleInputField.GetComponent<InputField>();
                    upgrade.OldFieldFemale = femaleInputField.GetComponent<InputField>();
                    
                    upgrade.NewFieldMale = this.UpgradeField(person, upgrade.OldFieldMale);                    
                    upgrade.NewFieldFemale = this.UpgradeField(person, upgrade.OldFieldFemale);

                    // the UI we are going to refresh is a sibling of the inputfield,
                    // so we find it from the parent's perspective.
                    Transform maleUITransform = upgrade.NewFieldMale.transform.parent.Find(k_morphUIPathMale);
                    Transform femaleUITransform = upgrade.NewFieldFemale.transform.parent.Find(k_morphUIPathFemale);
                    if (maleUITransform != null && femaleUITransform != null)
                    {
                        upgrade.UIMale = maleUITransform.GetComponent<GenerateDAZMorphsControlUI>();
                        upgrade.UIFemale = femaleUITransform.GetComponent<GenerateDAZMorphsControlUI>();
                        if (upgrade.UIMale == null || upgrade.UIFemale == null) this.generalError += " missing upgrade ui components !";
                    }
                    else this.generalError += " missing upgrade ui transforms !";
                }
                else
                {
                    this.generalError += " missing input fields !";
                    this.Log("Something went wrong and MorphFilterPlus couldn't find the normal input field. Did you do something?" +
                        " You changed something, didn't you? What have you done? WHAT DID YOU DO", error:true);                    
                }


                if (!this.hasError)
                {
                    // next, in order to actually do anything with the filter, we need to be able to
                    // access the morphs. Morphs are stored in components on child game objects of the person,
                    // one game object for each category of morph. These in turn are managed by 5 MorphBanks; 
                    // one for each gender's base and genital morphs, and an extra genital morph group for males.
                    // These 5 overarching MorphBanks are also on game objects-- they're the parents of the categories.
                    // they keep track of all of the morphs of their child subbanks.
                    // we are going to take advantage of that when we filter by using those parent morphbanks.

                    // rather than doing a GetComponent<> every time we filter, we're going to
                    // find and save the morph banks now as part of the upgrade object.
                    upgrade.MorphBanks = new DAZMorphBank[MorphFilterPlus.morphBankPaths.Length];
                    for (int i = 0; i < MorphFilterPlus.morphBankPaths.Length; i++)
                    {
                        upgrade.MorphBanks[i] = GetMorphBank(person, MorphFilterPlus.morphBankPaths[i]);
                        if (this.hasError) break;
                    }
                  
                }// end if load banks
            }// end if do upgrade 

            
            if (this.hasError)
            {                
                this.Log("MorphFilterPlus encountered errors: " + this.generalError, error:true);

                // if at any point we ran into a major error like a missing transform,
                // then this plugin is just not going to work, and we should give up and disable.
                this.StopWorkingAndSulk();
            }

        }

        
        /// <summary>
        /// Finds the child transform of the person according to a transform path,
        /// and fetches the morph bank component from it, if possible.
        /// Flags an error if it can't find the transform or morph bank.
        /// </summary>                
        private DAZMorphBank GetMorphBank(Atom person, string path)
        {
            // I use ret_ as a prefix for return values because I saw it in python and I like it.
            DAZMorphBank ret_bank = null;
            Transform bankTransform = person.transform.Find(path);
            if (bankTransform != null) ret_bank = bankTransform.GetComponent<DAZMorphBank>();

            // something's wrong if we don't find the bank. 
            // note the or-equals. If something is already wrong, we don't 
            // want to set it false if we did find the bank.
            if (ret_bank == null) this.generalError += $" missing morphbank at path '{path}' !";
            

            return ret_bank;
        }



        /// <summary>
        /// performs the legwork of duplicating and replacing an old field and intercepting its events,
        /// as well as applying a refreshing minty blue color. This is called within the UpgradePersonProcess.
        /// flags an error if it can't find important elements.
        /// </summary>        
        private InputField UpgradeField(Atom person, InputField oldField)
        {
            try
            {
                InputField ret_newField = Instantiate(oldField, oldField.transform.parent);

                // i'm not 100% sure if there's some layout stuff happening, so to preserve any
                // we make sure to plop the new field at the same spot in the hierarchy.
                ret_newField.transform.SetSiblingIndex(oldField.transform.GetSiblingIndex());

                // we don't want our duplicate field to send events to VAM's stuff,
                // so we scrub the events it might have...
                ret_newField.onValueChanged.RemoveAllListeners();
                ret_newField.onEndEdit.RemoveAllListeners();

                // ...and we add our own methods as listeners instead.
                ret_newField.onValueChanged.AddListener((value) => this.OnFilterChange(person, ret_newField));
                ret_newField.onEndEdit.AddListener((value) => this.OnFilterSubmit(person, ret_newField));

                // the clear filter buttun is actually a child of the filter.
                // this means it was duplicated along with the filter itself, which is nice.
                // we grab that and clear / intercept it just like the filter.
                var clearButton = ret_newField.GetComponentInChildren<Button>();
                if (clearButton != null)
                {
                    clearButton.onClick.RemoveAllListeners();
                    clearButton.onClick.AddListener(() => this.OnFilterClear(person, ret_newField));
                }
                else this.generalError += " missing clear/cancel button !";

                // add our signature MorphFilterBlue(TM)
                ret_newField.GetComponent<Image>().color = MorphFilterPlus.mintyFreshColor;

                // and we're done. wrap up by simply turning off the normal field
                oldField.gameObject.SetActive(false);                
                return ret_newField;
            }
            catch(Exception e)
            {
                this.generalError += " exception during UpgradeField(): " + e;
            }
            return null;
        }

        /// <summary>
        /// A general-use action coroutine we can add as a task
        /// to delay an arbitrary number of frames before doing something.        
        /// </summary>        
        /// <remarks>
        /// we use this to scan for people after loading a scene, for example,
        /// so that atoms are given a chance to initialize and any other plugins
        /// that might be messing with initialization are given a chance to do it.
        /// </remarks>
        private IEnumerator DelayedAction(int frameDelay, Action action)
        {
            int framesRemaining = frameDelay;
            while (!this.hasError && --framesRemaining > 0)
            {
                yield return null;
            }

            if (!this.hasError)
            {
                action();
            }
        }

        /// <summary>
        ///  We parse the text from the filter and apply it, 
        ///  IF live mode is active. This is us sneakily intercepting
        ///  the filter that would have gone to the morph UI.
        ///  We subscribed this to the input change event in our UpgradeField(..) method.
        /// </summary>        
        private void OnFilterChange(Atom person, InputField field)
        {            
            if (this.liveUpdateBoolJSON.val && (field.text.Length >= 3 || field.text.Length == 0))
            {                
                // to prevent aggressive filtering causing lag,
                // we only filter if we have at least 3 characters,
                // or are clearing the filter.
                this.ApplyFilter(person, field);
                
            }
        
        }


        /// <summary>
        /// Clears the text of the filter inputfield related to the button, which
        /// results in OnFilterChange() being called if there was any text to clear.
        /// Or, if our live update toggle is false, immediately restores morphs instead.
        /// We subscribed this to the clear button's OnClick event in our UpgradeField(..) method
        /// </summary>
        /// <param name="person"></param>
        /// <param name="field"></param>
        private void OnFilterClear(Atom person, InputField field)
        {            
            // setting the text will prompt OnFilterChange if it wasn't already empty,
            // so we don't have to do anything else here.
            field.text = "";
            if (this.submissionTask == null && !this.hasError)
            {                
                // if we haven't started to submit anything, then let's
                // restore everything and refresh the UI. If we HAVE submitted
                // something, then it's going to read the blank text and be fine anyway.
                this.RestoreMorphsForField(person, field);
                this.RefreshUIForField(person, field);
            }

            if (this.hasError)
            {
                this.StopWorkingAndSulk();
            }

        }



        private void OnFilterSubmit(Atom person, InputField field)
        {
            
            // we only use submission-based filtering if our live update toggle is off.
            if (!this.liveUpdateBoolJSON.val)
            {
                // here we must be clever.
                // the submit method is called when we hit enter OR click away from the field.
                // this means it's called when we click the Clear button (because we're clicking away).
                // if we don't account for that we'll end up applying the text filter 
                // and then clearing it immediately, which will both lag and possibly break some 
                // hidden complexity within VAM's morph or UI management.

                // so what we do instead is start a submission task where we wait until a 
                // few frames after the mouse has been released before applying the text of the inputfield.
                // this gives us a chance to clear it without applying it first.                
                this.submissionTask = this.SubmissionProcess(person, field);
            }
        }

        private IEnumerator SubmissionProcess(Atom person, InputField field)
        {
            // in case the mouse doesn't register as down this specific frame,
            // give it a sec.
            yield return null;
            while (Input.anyKey && !this.everythingIsTerrible)
            {
                // we use anykey to cover all input bases even though it's probably a mouse button.
                yield return null;
            }

            // once the button is released, give it a frame to publish its event
            // so that it has a chance to clear our input field.
            yield return null;

            this.ApplyFilter(person, field);
        }


        /// <summary>
        /// Fetches banks, builds the filters, and applies them to all the morphs,
        /// then refreshes the UI. This is the thing that occurs to actually
        /// make your morph ui show new stuff. It's called when an input field changes
        /// if live update is true, or after a submission event if live update is false.
        /// </summary>        
        private void ApplyFilter(Atom person, InputField field)
        {
            // this whole method structure is not very pretty or readable,
            // sort of has a function over form thing going on.
            // I'll probably prettify it later. For now I went ahead and added an early return
            // instead of nesting, as a readability choice, even though they are evil and I hate them.

            UpgradeProfile upgrade;
            if (!this.hasError && this.upgrades.TryGetValue(person, out upgrade))
            {                
                List<DAZMorphBank> banks = new List<DAZMorphBank>();

                // first we get the morph banks relevant to the field.
                if (field == upgrade.NewFieldFemale)
                {
                    banks.Add(upgrade.MorphBanks[BankIndex.FemaleChar]);
                    banks.Add(upgrade.MorphBanks[BankIndex.FemaleGen]);
                    
                }
                else if (field == upgrade.NewFieldMale)
                {
                    banks.Add(upgrade.MorphBanks[BankIndex.MaleChar]);
                    banks.Add(upgrade.MorphBanks[BankIndex.MaleGen]);
                    banks.Add(upgrade.MorphBanks[BankIndex.MaleGen2]);
                }
                else
                {
                    this.generalError += " ApplyFilter() unrecognized field !";
                }

                if (this.ignoredMorphs == null && !this.hasError)
                {
                    // if this is the first time we've tried applying a filter, we'll populate our ignored morphs cache.
                    // we waited this long because some morphs seem to be generated or hidden dynamically on selection,
                    // and we would have missed them in the upgrade step. here we know for sure the character has been selected.
                    this.ignoredMorphs = new HashSet<string>();
                    foreach (DAZMorphBank bank in upgrade.MorphBanks)
                    {
                        foreach (DAZMorph morph in bank.morphs)
                        {
                            // for some reason, autobehavior eyelid morphs don't get ignored in some cases.
                            // I think it's because they're generated dynamically or they start out visible and become invisible?
                            // whatever the case we can just add a manual ignore list to consider.
                            if (!morph.visible) this.ignoredMorphs.Add(morph.uid);
                        }
                    }
                    this.Log($"MorphFilterPlus is ignoring {this.ignoredMorphs.Count} morphs.");
                }

                // assuming we have banks to work with, we now build our filter list to apply to morphs.
                if (string.IsNullOrEmpty(field.text.Trim()))
                {                    
                    // if the filter is empty, we just restore all the morphs.
                    foreach (var bank in banks)
                    {
                        this.RestoreBank(bank);
                    }
                }
                else
                {                    

                    // we separate the string into each "OR" segment (%)
                    // and create a filter from each segment. 
                    // the filter constructor does the legwork of building constraints.
                    string[] segments = field.text.Split(new[] { '%' }, StringSplitOptions.RemoveEmptyEntries);
                    List<Filter> filters = new List<Filter>();
                    foreach (var segment in segments)
                    {
                        filters.Add(new Filter(segment.ToLower()));
                    }
                    
                    int spamcounter = 0;
                    foreach (DAZMorphBank bank in banks)
                    {
                        foreach (DAZMorph morph in bank.morphs)
                        {
                            // using continue to avoid nesting in a few places here.
                            if (this.ignoredMorphs.Contains(morph.uid))
                            {
                                continue;
                            }
                            
                            morph.visible = filters.Count == 0;
                            if (morph.visible) continue;

                            foreach (Filter filter in filters)
                            {
                                // we only need to match one of the OR filters, so as soon as we're visible, break.
                                morph.visible |= filter.CompareString(morph.lowerCaseResolvedDisplayName);
                                if (morph.visible) break;
                            }
                        }
                    }// end banksy loop
                }// end 'has text' condition

                // once we've modified all the relevant morphs' visibility,
                // we refresh the UI.
                this.RefreshUIForField(person, field);

            }// end method valid condition


            if (this.hasError)
            {
                this.StopWorkingAndSulk();
            }

        }// end method ApplyFilters()


        /// <summary>
        /// Given a person and a field, this gathers the relevant banks
        /// (such as the female character and genital banks for the female input field)
        /// and then restores all of their morphs.
        /// </summary>        
        private void RestoreMorphsForField(Atom person, InputField field)
        {
            UpgradeProfile upgrade;
            if (!this.hasError && this.upgrades.TryGetValue(person, out upgrade))
            {
                List<DAZMorphBank> banks = new List<DAZMorphBank>();
                if (field == upgrade.NewFieldFemale)
                {
                    banks.Add(upgrade.MorphBanks[BankIndex.FemaleChar]);
                    banks.Add(upgrade.MorphBanks[BankIndex.FemaleGen]);

                }
                else if (field == upgrade.NewFieldMale)
                {
                    banks.Add(upgrade.MorphBanks[BankIndex.MaleChar]);
                    banks.Add(upgrade.MorphBanks[BankIndex.MaleGen]);
                    banks.Add(upgrade.MorphBanks[BankIndex.MaleGen2]);
                }
                else
                {
                    this.generalError += " RestoreMorphsForField() unrecognized field !";
                }

                foreach (var bank in banks)
                {
                    this.RestoreBank(bank);
                }
            }
        }

        private void RefreshUIForField(Atom person, InputField field)
        {
            UpgradeProfile upgrade;
            if (!this.hasError && this.upgrades.TryGetValue(person, out upgrade))
            {
                if (field == upgrade.NewFieldFemale && upgrade.UIFemale != null)
                {
                    upgrade.UIFemale.ResyncMorphs();
                    upgrade.UIFemale.ForceCategoryRefresh();
                }
                else if (field == upgrade.NewFieldMale && upgrade.UIMale != null)
                {
                    upgrade.UIMale.ResyncMorphs();
                    upgrade.UIMale.ForceCategoryRefresh();
                }
                else
                {
                    this.generalError += " RefreshUIForField() unrecognized field or missing UI";
                }
            }
        }

        /// <summary>
        /// Called by Unity every frame automatically as part of the MonoBehavior API
        /// we inherit. We use it to step through our coroutines each frame. 
        /// Update has overhead so it's a tiny bit wasteful for this. I don't want to use it.
        /// but I really REALLY don't like StartCoroutine, so I'm going to anyway for now.
        /// </summary>
        private void Update()
        {            
            for (int i = 0; i < this.tasks.Count; i++)
            {
                if (!this.tasks[i].MoveNext())
                {
                    this.tasks.RemoveAt(i);
                    i--;
                }
            }

            // if we are not in live update mode,
            // then we update a submission process if we have one.
            if (this.liveUpdateBoolJSON != null && !this.liveUpdateBoolJSON.val && this.submissionTask != null)
            {
                if (!this.submissionTask.MoveNext()) this.submissionTask = null;
            }

        }


        /// <summary>
        /// Responds to newly created atoms by checking if they're a person
        /// and upgrading them if so. This is one of our event subscriptions in OnEnable()
        /// </summary>        
        private void OnAtomAdded(Atom atom)
        {
            if (atom.type == k_Person)
            {
                this.UpgradePerson(atom);
            }
        }

        /// <summary>
        /// Responds to the removal of an atom by checking if they're a person
        /// and reverting their upgrade if so. This is one of our event subscriptions in OnEnable()
        /// </summary>
        /// <param name="atom"></param>
        private void OnAtomRemoved(Atom atom)
        {
            if (atom != null && atom.type == k_Person)
            {                
                this.RevertUpgrade(atom);
            }
        }

        /// <summary>
        /// Responds to the VAM scene being loaded by scanning for people present,
        /// after a brief delay to allow any weird atom things to be settled.
        /// </summary>
        private void OnSceneLoaded()
        {
            IEnumerator delayedScan = this.DelayedAction(frameDelay:5, action:this.ScanForPeople);
            this.tasks.Add(delayedScan);
        }


        /// <summary>
        /// Responds to a VAM subscene being loaded by scanning for new people,
        /// after a brief delay to allow any weird atom things to be settled.
        /// </summary>        
        private void OnSubSceneLoaded(SubScene scene)
        {
            IEnumerator delayedScan = this.DelayedAction(frameDelay: 5, action: this.ScanForPeople);
            this.tasks.Add(delayedScan);
        }

        /// <summary>
        /// Removes the filter upgrade applied to a specific person,
        /// resulting in restoration of the original filter UI
        /// </summary>        
        public void RevertUpgrade(Atom person)
        {
            if (this.upgrades.ContainsKey(person))
            {
                this.Log($"MorphFilterPlus reverting person '{(person != null ? person.name: "NULL")}'");
                this.RestoreMorphs(person);
                this.upgrades[person].Revert();
                this.upgrades.Remove(person);
            }
            
        }


        /// <summary>
        /// Restores all banks of the person's upgrade profile,
        /// revering all their morphs to visible (except the ignored ones)
        /// </summary>        
        private void RestoreMorphs(Atom person)
        {
            if (this.upgrades[person].MorphBanks != null)
            {
                foreach (DAZMorphBank bank in this.upgrades[person].MorphBanks)
                {
                    if (bank != null) this.RestoreBank(bank);
                }
            }

            // we have to resync the morphs, otherwise 
            // VAM won't know to restore them on character change.
            if (this.upgrades[person].UIMale != null)
            {
                this.upgrades[person].UIMale.ResyncMorphs();
            }
            if (this.upgrades[person].UIFemale != null)
            {
                this.upgrades[person].UIFemale.ResyncMorphs();
            }

        }

        /// <summary>
        /// Restores all morphs of a specific morph bank,
        /// reverting them to visible (except the ignored ones)
        /// </summary>        
        private void RestoreBank(DAZMorphBank bank)
        {        
            foreach (DAZMorph morph in bank.morphs)
            {
                if (morph != null) morph.visible = !this.ignoredMorphs.Contains(morph.uid);

                // some invisible morphs get modified by appearances or drivers or other means.
                // and sometimes they seem to get mixed up with non-invisible morphs with the same 
                // display name. I'm still trying to find more out, but in the mean time
                // when we revert a morph bank, we'll just reset the invis morphs uids,
                // to make sure no changes are applied that you can't see to unapply.
                if (!morph.visible) morph.Reset();
            }
        }

        /// <summary>
        /// Sets the flag implicating that something is critically wrong,
        /// and disables the script. This is called when we encounter an issue
        /// that would prevent the plugin from working, such as missing transforms.
        /// Since it disables itself, it results in the reversion of any existing upgrades.
        /// </summary>
        private void StopWorkingAndSulk()
        {
            if (!this.everythingIsTerrible)
            {

                this.Log($"MorphFilterPlus is stopping due to errors: {this.generalError} \nThings didn't go as planned, and MorphFilterPlus has stopped working for now. " +
                    "You can try restarting or reloading it if you want.", error: true);

                this.everythingIsTerrible = true;
                this.generalError = "";

                // we don't want to just disable the script, we want to show it as disabled
                // in VAM's plugin UI by turning off the toggle. Doing so will result
                // in a callback that disables it, calls OnDisable, and reverts our upgrades as desired.
                // however if there is a nicer looking way to do this, let me (GossamerVR) know.                
                if (this.enabledJSON != null && this.enabledJSON.toggle != null)
                {
                    this.enabledJSON.toggle.isOn = false;
                }

            }
        }



        private void Log(bool condition, string str, bool error = false)
        {
            if (condition) this.Log(str);
        }

        private void Log(string str, bool error = false)
        {
            if (error) SuperController.LogError(str);
            else SuperController.LogMessage(str);
        }


        /// <summary>
        /// stores references to the new and old objects after an upgrade 
        /// takes place, and provides a method to revert the changes.
        /// </summary>
        private class UpgradeProfile
        {            
            public InputField OldFieldFemale;
            public InputField OldFieldMale;

            public InputField NewFieldFemale;
            public InputField NewFieldMale;

            public GenerateDAZMorphsControlUI UIFemale;
            public GenerateDAZMorphsControlUI UIMale;
            
            public DAZMorphBank[] MorphBanks;

            public void Revert()
            {                
                if (this.OldFieldFemale != null) this.OldFieldFemale.gameObject.SetActive(true);
                if (this.OldFieldMale != null) this.OldFieldMale.gameObject.SetActive(true);

                if (this.NewFieldFemale != null) Destroy(this.NewFieldFemale.gameObject);
                if (this.NewFieldMale != null) Destroy(this.NewFieldMale.gameObject);

                // while not strictly necessary, we want to make sure references are released
                // to avoid memory bloat, in case VAM or Unity or a plugin maintains a reference
                // to this object somewhere, somehow.
                this.OldFieldFemale = null;
                this.OldFieldMale = null;

                this.UIMale = null;
                this.UIFemale = null;

                this.MorphBanks = null;

            }
        }

    }
}